/** 
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#include "wlpdstm.h"

#include "tid.h"
#include "memory.h"
#include "transaction.h"
#include "constants.h"

///////////////////////////
// basic interface start //
///////////////////////////	

void wlpdstm::global_init() {
	CurrentTransaction::GlobalInit();
}

void wlpdstm::thread_init() {
	CurrentTransaction::ThreadInit();
}

void wlpdstm::start_tx() {
	CurrentTransaction::TxStart();
}

void wlpdstm::start_tx(int lexical_tx_id) {
	CurrentTransaction::TxStart(lexical_tx_id);
}

LONG_JMP_BUF *wlpdstm::get_long_jmp_buf() {
	return CurrentTransaction::GetLongJmpBuf();
}

void wlpdstm::commit_tx() {
	CurrentTransaction::TxCommit();
}

void wlpdstm::abort_tx() {
	CurrentTransaction::TxAbort();
}

void wlpdstm::restart_tx() {
	CurrentTransaction::TxRestart();
}

void wlpdstm::write_word(Word *address, Word value) {
	CurrentTransaction::WriteWord(address, value);
}

#ifdef SUPPORT_LOCAL_WRITES
void wlpdstm::write_word_local(Word *address, Word value) {
	CurrentTransaction::WriteWordLocal(address, value);
}
#endif /* SUPPORT_LOCAL_WRITES */

wlpdstm::Word wlpdstm::read_word(Word *address) {
	return CurrentTransaction::ReadWord(address);
}

void wlpdstm::tx_free(void *ptr, size_t size) {
	CurrentTransaction::TxFree(ptr, size);
}

void *wlpdstm::tx_malloc(size_t size) {
	return CurrentTransaction::TxMalloc(size);
}

wlpdstm::tx_desc *wlpdstm::get_tx_desc() {
	return (tx_desc *)CurrentTransaction::Get();
}

void wlpdstm::start_tx(tx_desc *tx) {
	((Transaction *)tx)->TxStart();
}

LONG_JMP_BUF *wlpdstm::get_long_jmp_buf(tx_desc *tx) {
	return &((Transaction *)tx)->start_buf;
}

void wlpdstm::start_tx(tx_desc *tx, int lexical_tx_id) {
	((Transaction *)tx)->TxStart(lexical_tx_id);
}

void wlpdstm::commit_tx(tx_desc *tx) {
	((Transaction *)tx)->TxCommit();
}

void wlpdstm::restart_tx(tx_desc *tx) {
	((Transaction *)tx)->TxRestart();
}

void wlpdstm::abort_tx(tx_desc *tx) {
	((Transaction *)tx)->TxAbort();
}

wlpdstm::Word wlpdstm::read_word(tx_desc *tx, Word *address) {
	return ((Transaction *)tx)->ReadWord(address);
}

void wlpdstm::write_word(tx_desc *tx, Word *address, Word value) {
	((Transaction *)tx)->WriteWord(address, value);
}

#ifdef SUPPORT_LOCAL_WRITES
void wlpdstm::write_word_local(tx_desc *tx, Word *address, Word value) {
	((Transaction *)tx)->WriteWordLocal(address, value);
}
#endif /* SUPPORT_LOCAL_WRITES */

void wlpdstm::tx_free(tx_desc *tx, void *ptr, size_t size) {
	((Transaction *)tx)->TxFree(ptr, size);
}

void *wlpdstm::tx_malloc(tx_desc *tx, size_t size) {
	return ((Transaction *)tx)->TxMalloc(size);
}

void wlpdstm::print_stats() {
	Transaction::PrintStatistics();
}

void *wlpdstm::s_malloc(size_t size) {
	return wlpdstm::MemoryManager::Malloc(size);
}

void wlpdstm::s_free(void *ptr) {
	wlpdstm::MemoryManager::Free(ptr);
}

/////////////////////////
// basic interface end //
/////////////////////////	


//////////////////////////////
// extended interface start //
//////////////////////////////

typedef union word_64_u {
	uint64_t b64;
	uint32_t b32[2];
	float f[2];
} word_64;

#ifdef X86
uint32_t wlpdstm::read_32(uint32_t *address) {
	return CurrentTransaction::ReadWord((Word *)address);
}

uint32_t wlpdstm::read_32(tx_desc *tx, uint32_t *address) {
	return ((Transaction *)tx)->ReadWord((Word *)address);
}

float wlpdstm::read_float(float *address) {
	word_64 val;
	val.b32[0] = CurrentTransaction::ReadWord((Word *)address);
	return val.f[0];
}

float wlpdstm::read_float(tx_desc *tx, float *address) {
	word_64 val;
	val.b32[0] = ((Transaction *)tx)->ReadWord((Word *)address);
	return val.f[0];
}

void wlpdstm::write_32(uint32_t *address, uint32_t value) {
	CurrentTransaction::WriteWord((Word *)address, value);
}

void wlpdstm::write_32(tx_desc *tx, uint32_t *address, uint32_t value) {
	((Transaction *)tx)->WriteWord((Word *)address, value);
}

void wlpdstm::write_float(float *address, float value) {
	word_64 val;
	val.f[0] = value;
	CurrentTransaction::WriteWord((Word *)address, val.b32[0]);
}

void wlpdstm::write_float(tx_desc *tx, float *address, float value) {
	word_64 val;
	val.f[0] = value;
	((Transaction *)tx)->WriteWord((Word *)address, val.b32[0]);
}

#elif defined X86_64

#define MASK_64_32 0x7l
#define SHIFT_32 2

#define WORD32_POS(address)     (((uintptr_t)address & MASK_64_32) >> SHIFT_32)
#define WORD32_ADDRESS(address) ((Word *)((uintptr_t)address & ~MASK_64_32))

uint32_t wlpdstm::read_32(uint32_t *address) {
	word_64 value;
	value.b64 = (uint64_t)CurrentTransaction::ReadWord(WORD32_ADDRESS(address));
	return value.b32[WORD32_POS(address)];
}

uint32_t wlpdstm::read_32(tx_desc *tx, uint32_t *address) {
	word_64 value;
	value.b64 = (uint64_t)((Transaction *)tx)->ReadWord(WORD32_ADDRESS(address));
	return value.b32[WORD32_POS(address)];	
}

float wlpdstm::read_float(float *address) {
	word_64 value;
	value.b64 = (uint64_t)CurrentTransaction::ReadWord(WORD32_ADDRESS(address));
	return value.f[WORD32_POS(address)];
}

float wlpdstm::read_float(tx_desc *tx, float *address) {
	word_64 value;
	value.b64 = (uint64_t)((Transaction *)tx)->ReadWord(WORD32_ADDRESS(address));
	return value.f[WORD32_POS(address)];
}

#define PREPARE_32_MASK_VAL(address, value, field) \
    unsigned pos = WORD32_POS(address);            \
    word_64 mask;                                  \
    mask.b64 = 0;                                  \
    mask.b32[pos] = ~0;                            \
    word_64 val;                                   \
    val.field[pos] = value


void wlpdstm::write_32(uint32_t *address, uint32_t value) {
	PREPARE_32_MASK_VAL(address, value, b32);
	CurrentTransaction::WriteWordMasked(WORD32_ADDRESS(address), val.b64, mask.b64);
}

void wlpdstm::write_32(tx_desc *tx, uint32_t *address, uint32_t value) {
	PREPARE_32_MASK_VAL(address, value, b32);
	((Transaction *)tx)->WriteWordMasked(WORD32_ADDRESS(address), val.b64, mask.b64);	
}

void wlpdstm::write_float(float *address, float value) {
	PREPARE_32_MASK_VAL(address, value, f);
	CurrentTransaction::WriteWordMasked(WORD32_ADDRESS(address), val.b64, mask.b64);
}

void wlpdstm::write_float(tx_desc *tx, float *address, float value) {
	PREPARE_32_MASK_VAL(address, value, f);
	((Transaction *)tx)->WriteWordMasked(WORD32_ADDRESS(address), val.b64, mask.b64);
}

#endif /* X86 */


//////////////////////////////
// extended interface end //
//////////////////////////////
