/**
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef WLPDSTM_WLPDSTM_H_
#define WLPDSTM_WLPDSTM_H_

#include <cstdlib>
#include <stdint.h>

#include "common/word.h"
#include "common/jmp.h"

namespace wlpdstm {

	///////////////////////////
	// basic interface start //
	///////////////////////////	
	
	typedef void *tx_desc;

	// initialization
	void global_init();

	void thread_init();

	// start/end tx
	void start_tx();
	void start_tx(int lexical_tx_id);

	LONG_JMP_BUF *get_long_jmp_buf();

	void commit_tx();

	void abort_tx();

	void restart_tx();

	// read/write word
	Word read_word(Word *address);

	void write_word(Word *address, Word value);

#ifdef SUPPORT_LOCAL_WRITES
	void write_word_local(Word *address, Word value);
#endif /* SUPPORT_LOCAL_WRITES */

	// memory management
	void *tx_malloc(size_t size);

	void tx_free(void *ptr, size_t size);

	///////////////////////////////////////////////////////
	// a separate set of methods if current tx is cached //
	///////////////////////////////////////////////////////

	tx_desc *get_tx_desc();

	// start/end tx
	void start_tx(tx_desc *tx);
	void start_tx(tx_desc *tx, int lexical_tx_id);

	LONG_JMP_BUF *get_long_jmp_buf(tx_desc *tx);

	void commit_tx(tx_desc *tx);

	void abort_tx(tx_desc *tx);

	void restart_tx(tx_desc *tx);

	// read/write word
	Word read_word(tx_desc *tx, Word *address);

	void write_word(tx_desc *tx, Word *address, Word value);

#ifdef SUPPORT_LOCAL_WRITES
	void write_word_local(tx_desc *tx, Word *address, Word value);
#endif /* SUPPORT_LOCAL_WRITES */

	// memory management
	void *tx_malloc(tx_desc *tx, size_t size);

	void tx_free(tx_desc *tx, void *ptr, size_t size);

	void print_stats();

	// use for non-tx code to be able to simply switch
	// the whole memory management scheme
	void *s_malloc(size_t size);
	void s_free(void *ptr);

	/////////////////////////
	// basic interface end //
	/////////////////////////	

	//////////////////////////////
	// extended interface start //
	//////////////////////////////

	uint32_t read_32(uint32_t *address);
	uint32_t read_32(tx_desc *tx, uint32_t *address);
	float read_float(float *address);
	float read_float(tx_desc *tx, float *address);

	void write_32(uint32_t *address, uint32_t value);
	void write_32(tx_desc *tx, uint32_t *address, uint32_t value);
	void write_float(float *address, float value);
	void write_float(tx_desc *tx, float *address, float value);

	////////////////////////////
	// extended interface end //
	////////////////////////////	
}

#endif // WLPDSTM_WLPDSTM_H_
